"""
pathselection.py

Created by Nick on 2011-06-27.
"""

from cadnano2.views import styles
from math import floor
import cadnano2.util as util

# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['Qt', 'QPointF', 'QEvent', 'QRectF'])
util.qtWrapImport('QtGui', globals(), ['QPen', 'QBrush', 'QColor',
                                       'QPainterPath'])
util.qtWrapImport('QtWidgets', globals(), [  # 'qApp',
                                           'QGraphicsItem',
                                           'QGraphicsItemGroup',
                                           'QGraphicsPathItem'])


class SelectionItemGroup(QGraphicsItemGroup):
    """
    SelectionItemGroup
    """
    def __init__(self, boxtype, constraint='y', parent=None):
        super(SelectionItemGroup, self).__init__(parent)
        self._viewroot = parent
        self.setFiltersChildEvents(True)

        # LOOK at Qt Source for deprecated code to replace this behavior
        # self.setHandlesChildEvents(True) # commented out NC

        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsFocusable)  # for keyPressEvents
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemHasNoContents)

        self._rect = QRectF()
        self._pen = QPen(styles.bluestroke, styles.PATH_SELECTBOX_STROKE_WIDTH)

        self.selectionbox = boxtype(self)

        self._dragEnable = False
        self._dragged = False
        self._baseClick = 0  # tri-state counter to enable deselection on click away

        self._r0 = 0  # save original mousedown
        self._r = 0  # latest position for moving

        # self._lastKid = 0

        # this keeps track of mousePressEvents within the class
        # to aid in intellignetly removing items from the group
        self._addedToPressList = False

        self._pendingToAddDict = {}

        if constraint == 'y':
            self.getR = self.selectionbox.getY
            self.translateR = self.selectionbox.translateY
        else:
            self.getR = self.selectionbox.getX
            self.translateR = self.selectionbox.translateX

        self._normalSelect = True
        self._instantAdd = 0

        self.setZValue(styles.ZPATHSELECTION)
    # end def

    # def paint(self, painter, option, widget):
    #     painter.drawRect(self.boundingRect())
    # # end def

    def pendToAdd(self, item):
        self._pendingToAddDict[item] = True
    # end def

    def isPending(self, item):
        return item in self._pendingToAddDict
    # end def
    
    def document(self):
        return self._viewroot.document()
    # end def
    
    def pendToRemove(self, item):
        if item in self._pendingToAddDict:
            del self._pendingToAddDict[item]
    # end def

    def setNormalSelect(self, boolVal):
        self._normalSelect = boolVal
    # end def

    def isNormalSelect(self):
        return self._normalSelect
    # end def

    def processPendingToAddList(self):
        """
        Adds to the local selection and the document if required
        """
        doc = self.document()

        # print "instant add is 1 from process pending"
        if len(self._pendingToAddDict) > 0:
            for item in self._pendingToAddDict:
                # print "just checking1", item, item.group(), item.parentItem()
                self.addToGroup(item)
                item.modelSelect(doc)
            # end for
            self._pendingToAddDict = {}
            doc.updateSelection()
    # end def

    def resetSelection(self):
        self._pendingToAddDict = {}
        self._addedToPressList = False
        self.clearSelection(False)
        self.setSelectionLock(None)
        self.selectionbox.setParentItem(self._viewroot)
        self.setParentItem(self._viewroot)
    # end def

    def selectionLock(self):
        return self._viewroot.selectionLock()
    # end def

    def setSelectionLock(self, selectionGroup):
        self._viewroot.setSelectionLock(selectionGroup)
    # end def

    # def paint(self, painter, option, widget=None):
    #     # painter.setBrush(QBrush(QColor(255,128,255,128)))
    #     painter.setPen(QPen(styles.redstroke))
    #     painter.drawRect(self.boundingRect())
    #     pass
    # # end def

    def keyPressEvent(self, event):
        """
        Must intercept invalid input events.  Make changes here
        """
        key = event.key()
        if key in [Qt.Key.Key_Backspace, Qt.Key.Key_Delete]:
            self.selectionbox.deleteSelection()
            # self.document().deleteSelection()
            self.clearSelection(False)
            return QGraphicsItemGroup.keyPressEvent(self, event)
        else:
            return QGraphicsItemGroup.keyPressEvent(self, event)
    # end def

    def mousePressEvent(self, event):
        # self.show()
        # print "instant add is 1 from MousePress"
        if event.button() != Qt.MouseButton.LeftButton:
            return QGraphicsItemGroup.mousePressEvent(self, event)
        else:
            self._dragEnable = True

            # required to get the itemChanged event to work
            # correctly for this
            self.setSelected(True)

            # self.selectionbox.resetTransform()
            self.selectionbox.resetPosition()
            self.selectionbox.refreshPath()

            # self.selectionbox.resetTransform()
            self.selectionbox.resetPosition()
            self.selectionbox.show()

            # for some reason we need to skip the first mouseMoveEvent
            self._dragged = False

            if self._addedToPressList == False:
                self._addedToPressList = True
                self.scene().views()[0].addToPressList(self)
            return QGraphicsItemGroup.mousePressEvent(self, event)
    # end def

    def mouseMoveEvent(self, event):
        if self._dragEnable == True:
            # map the item to the scene coordinates
            # to help keep coordinates uniform
            rf = self.getR(self.mapFromScene(QPointF(event.scenePos())))
            # for some reason we need to skip the first mouseMoveEvent
            if self._dragged == False:
                self._dragged = True
                self._r0 = rf
            # end if
            else:
                delta = self.selectionbox.delta(rf, self._r0)
                self.translateR(delta)
            # end else
            self._r = rf
        # end if
        else:
            QGraphicsItemGroup.mouseMoveEvent(self, event)
        # end else
    # end def

    def customMouseRelease(self, event):
        """docstring for customMouseRelease"""
        self.selectionbox.hide()
        self.selectionbox.resetTransform()
        self._dragEnable = False
        # now do stuff
        if not (self._r0 == 0 and self._r == 0):
            self.selectionbox.processSelectedItems(self._r0, self._r)
        # end if
        self._r0 = 0  # reset
        self._r = 0  # reset
        self.setFocus() # needed to get keyPresses post a move
        self._addedToPressList = False
    # end def

    def clearSelection(self, value):
        if value == False:
            self.selectionbox.hide()
            self.selectionbox.resetPosition()
            self.removeSelectedItems()
            self._viewroot.setSelectionLock(None)
            self.clearFocus()  # this is to disable delete keyPressEvents
            self.prepareGeometryChange()
            self._rect.setWidth(0)
            # self._rect = QRectF()
        # end if
        else:
            self.setFocus()  # this is to get delete keyPressEvents
        self.update(self.boundingRect())
    # end def

    def itemChange(self, change, value):
        """docstring for itemChange"""
        if change == QGraphicsItem.GraphicsItemChange.ItemSelectedChange:
            if value == False:
                self.clearSelection(False)
                return False
            else:
                return True
        elif change == QGraphicsItem.GraphicsItemChange.ItemChildAddedChange:
            if self._addedToPressList == False:
                # print "kid added"
                self.setFocus()  # this is to get delete keyPressEvents
                self.setParentItem(self.selectionbox.boxParent())
                self._addedToPressList = True
                self.scene().views()[0].addToPressList(self)
            return
        return QGraphicsItemGroup.itemChange(self, change, value)
    # end def

    def removeChild(self, child):
        """
        remove only the child and ask it to
        restore it's original parent
        """
        doc = self.document()
        tPos = child.scenePos()
        self.removeFromGroup(child)
        child.modelDeselect(doc)
    # end def

    def removeSelectedItems(self):
        """docstring for removeSelectedItems"""
        doc = self.document()
        for item in self.childItems():
            self.removeFromGroup(item)
            item.modelDeselect(doc)
        # end for
        doc.updateSelection()
    # end def

    def setBoundingRect(self, rect):
        self.prepareGeometryChange()
        self._rect = rect
    # end def

    def boundingRect(self):
        return self._rect

# end class


class VirtualHelixHandleSelectionBox(QGraphicsPathItem):
    """
    docstring for VirtualHelixHandleSelectionBox
    """
    _helixHeight = styles.PATH_HELIX_HEIGHT + styles.PATH_HELIX_PADDING
    _radius = styles.VIRTUALHELIXHANDLEITEM_RADIUS
    _penWidth = styles.SLICE_HELIX_HILIGHT_WIDTH
    _boxPen = QPen(styles.bluestroke, _penWidth)

    def __init__(self, itemGroup):
        """
        The itemGroup.parentItem() is expected to be a partItem
        """
        super(VirtualHelixHandleSelectionBox, self).__init__(
                                                        itemGroup.parentItem())
        self._itemGroup = itemGroup
        self._rect = itemGroup.boundingRect()
        self.hide()
        self.setPen(self._boxPen)
        self.setZValue(styles.ZPATHSELECTION)
        self._bounds = None
        self._pos0 = QPointF()
    # end def

    def getY(self, pos):
        pos = self._itemGroup.mapToScene(QPointF(pos))
        return pos.y()
    # end def

    def translateY(self, delta):
        self.setY(delta)
     # end def

    def refreshPath(self):
        self.prepareGeometryChange()
        self.setPath(self.painterPath())
        self._pos0 = self.pos()
    # end def

    def painterPath(self):
        iG = self._itemGroup
        # the childrenBoundingRect is necessary to get this to work
        rect = self.mapRectFromItem(iG, iG.childrenBoundingRect())
        radius = self._radius

        path = QPainterPath()
        path.addRoundedRect(rect, radius, radius)
        path.moveTo(rect.right(),\
                         rect.center().y())
        path.lineTo(rect.right() + radius / 2,\
                         rect.center().y())
        return path
    # end def

    def processSelectedItems(self, rStart, rEnd):
        """docstring for processSelectedItems"""
        margin = styles.VIRTUALHELIXHANDLEITEM_RADIUS
        delta = (rEnd - rStart)  # r delta
        midHeight = (self.boundingRect().height()) / 2 - margin
        helixHeight = self._helixHeight

        if abs(delta) < midHeight:  # move is too short for reordering
            return
        if delta > 0:  # moved down, delta is positive
            indexDelta = int((delta - midHeight) / helixHeight)
        else:  # moved up, delta is negative
            indexDelta = int((delta + midHeight) / helixHeight)
        # sort on y to determine the extremes of the selection group
        items = sorted(self._itemGroup.childItems(), key=lambda vhhi: vhhi.y())
        partItem = items[0].partItem()
        partItem.reorderHelices(items[0].number(),\
                                items[-1].number(),\
                                indexDelta)
        partItem.updateStatusBar("")
    # end def

    def boxParent(self):
        temp = self._itemGroup.childItems()[0].partItem()
        self.setParentItem(temp)
        return temp
    # end def
    
    def deleteSelection(self):
        """
        Delete selection operates outside of the documents a virtual helices
        are not actually selected in the model
        """
        vHelices = [vhh.virtualHelix() for vhh in self._itemGroup.childItems()]
        uS = self._itemGroup.document().undoStack()
        uS.beginMacro("delete Virtual Helices")
        for vh in vHelices:
            vh.remove()
        uS.endMacro()
    # end def

    def bounds(self):
        return self._bounds
    # end def

    def delta(self, yf, y0):
        return yf - y0
    # end def

    def resetPosition(self):
        self.setPos(self._pos0)
    # end def
# end class


class EndpointHandleSelectionBox(QGraphicsPathItem):
    _penWidth = styles.SLICE_HELIX_HILIGHT_WIDTH
    _boxPen = QPen(styles.selected_color, _penWidth)
    _baseWidth = styles.PATH_BASE_WIDTH

    def __init__(self, itemGroup):
        """
        The itemGroup.parentItem() is expected to be a partItem
        """
        super(EndpointHandleSelectionBox, self).__init__(
                                                    itemGroup.parentItem())
        self._itemGroup = itemGroup
        self._rect = itemGroup.boundingRect()
        self.hide()
        self.setPen(self._boxPen)
        self.setZValue(styles.ZPATHSELECTION)
        self._bounds = (0, 0)
        self._pos0 = QPointF()
    # end def

    def getX(self, pos):
        return pos.x()
    # end def

    def translateX(self, delta):
        children = self._itemGroup.childItems()
        if children:
            pI = children[0].partItem()
            str = "+%d" % delta if delta >= 0 else "%d" % delta
            pI.updateStatusBar(str)
        self.setX(self._baseWidth * delta)
    # end def

    def resetPosition(self):
        self.setPos(self._pos0)

    def delta(self, xf, x0):
        boundL, boundH = self._bounds
        delta = int(floor((xf - x0) / self._baseWidth))
        if delta > 0 and delta > boundH:
            delta = boundH
        elif delta < 0 and abs(delta) > boundL:
            delta = -boundL
        return delta

    def refreshPath(self):
        tempLow, tempHigh = \
                    self._itemGroup._viewroot.document().getSelectionBounds()
        self._bounds = (tempLow, tempHigh)
        self.prepareGeometryChange()
        self.setPath(self.painterPath())
        self._pos0 = self.pos()
    # end def

    def painterPath(self):
        bw = self._baseWidth
        iG = self._itemGroup
        # the childrenBoundingRect is necessary to get this to work
        rectIG = iG.childrenBoundingRect()
        rect = self.mapRectFromItem(iG, rectIG)
        if rect.width() < bw:
            rect.adjust(-bw / 4, 0, bw / 2, 0)
        path = QPainterPath()
        path.addRect(rect)
        self._itemGroup.setBoundingRect(rectIG)

        # path.addRoundedRect(rect, radius, radius)
        # path.moveTo(rect.right(),\
        #                  rect.center().y())
        # path.lineTo(rect.right() + radius / 2,\
        #                  rect.center().y())
        return path
    # end def

    def processSelectedItems(self, rStart, rEnd):
        """docstring for processSelectedItems"""
        delta = self.delta(rEnd, rStart)
        self._itemGroup._viewroot.document().resizeSelection(delta)
    # end def
    
    def deleteSelection(self):
        self._itemGroup.document().deleteSelection()

    def boxParent(self):
        temp = self._itemGroup.childItems()[0].partItem()
        self.setParentItem(temp)
        return temp
    # end def

    def bounds(self):
        return self._bounds
    # end def
# end class



import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtCore', globals(), ['pyqtSignal', 'QObject'])
util.qtWrapImport('QtGui', globals(), [ 'QActionGroup'])


class SliceToolManager(QObject):
    """Manages interactions between the slice widgets/UI and the model."""
    def __init__(self, win):
        """
        We store mainWindow because a controller's got to have
        references to both the layer above (UI) and the layer below (model)
        """
        super(SliceToolManager, self).__init__()
        self._window = win
        self._connectWindowSignalsToSelf()

    ### SIGNALS ###
    activeSliceSetToFirstIndexSignal = pyqtSignal()
    activeSliceSetToLastIndexSignal = pyqtSignal()
    activePartRenumber = pyqtSignal()

    ### SLOTS ###
    def activeSliceFirstSlot(self):
        """
        Use a signal to notify the ActiveSliceHandle to move. A signal is used
        because the SliceToolManager must be instantiated first, and the
        ActiveSliceHandle can later subscribe.
        """
        part = self._window.selectedPart()
        if part != None:
            part.setActiveBaseIndex(0)

    def activeSliceLastSlot(self):
        part = self._window.selectedPart()
        if part != None:
            part.setActiveBaseIndex(part.maxBaseIdx()-1)

    ### METHODS ###
    def _connectWindowSignalsToSelf(self):
        """This method serves to group all the signal & slot connections
        made by SliceToolManager"""
        self._window.actionSliceFirst.triggered.connect(self.activeSliceFirstSlot)
        self._window.actionSliceLast.triggered.connect(self.activeSliceLastSlot)



###############################################################################
#
# Copyright 2011 Autodesk, Inc.  All rights reserved.
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
###############################################################################


import maya.cmds as cmds
import maya.mel as mel

# Original States
hiddenElements = []
elementsToHide = ["toolBar", "dockControl"]
mainMenuBarVisible = None
gridVisible = None

modelEditors = []
backgroundColors = []
"""
Hideable Stuff:
    UIElements
    Grid
    MainMenuBar
"""

def simplifyUI():
    hideUIElements()
    hideMainMenuBar()
    hideGrid()
    #mel.eval("toggleModelEditorBarsInAllPanels 0;")

def restoreUI():
    restoreUIElements()
    restoreMainMenuBar()
    restoreGrid()
    #mel.eval("toggleModelEditorBarsInAllPanels 1;")

def setViewportQuality():
    global modelEditors
    global backgroundColors
    backgroundColors.append(cmds.displayRGBColor('background',q=True))
    backgroundColors.append(cmds.displayRGBColor('backgroundTop',q=True))
    backgroundColors.append(cmds.displayRGBColor('backgroundBottom',q=True))
    cmds.displayRGBColor('background', 1, 1, 1)
    cmds.displayRGBColor('backgroundTop', 0.762112, 0.87892, 1)
    cmds.displayRGBColor('backgroundBottom', 1, 1, 1)

    for i in cmds.lsUI(panels=True):
        if cmds.modelEditor(i, query=True, exists=True):
            sts = cmds.modelEditor(i, query=True, stateString=True)
            sts = sts.replace("$editorName", i)
            modelEditors.append(sts)
            cmds.modelEditor(i, edit=True, displayAppearance="smoothShaded", lineWidth=2)
                                 #, rendererName="hwRender_OpenGL_Renderer")

def restoreViewportQuality():
    global modelEditors
    global backgroundColors
    bc = backgroundColors
    cmds.displayRGBColor('background', bc[0][0], bc[0][1], bc[0][2])
    cmds.displayRGBColor('backgroundTop', bc[1][0], bc[1][1], bc[1][2])
    cmds.displayRGBColor('backgroundBottom', bc[2][0], bc[2][1], bc[2][2])

    for e in modelEditors:
        mel.eval(e)

def hideMainMenuBar():
    global mainMenuBarVisible
    mainMenuBarVisible = cmds.optionVar(q="mainWindowMenubarVis")
    #mel.eval("setMainMenubarVisible 0;")

def restoreMainMenuBar():
    global mainMenuBarVisible
    #mel.eval("setMainMenubarVisible " + str(mainMenuBarVisible) + ";")

def hideUIElements():
    global hiddenElements
    global elementsToHide
    for i in cmds.lsUI(ctl=True):
        for e in elementsToHide:
            if i.find(e) != -1 and cmds.control(i, q=True, visible=True):
                hiddenElements.append(i)
                #print "hiding... " + i
                cmds.control(i, e=True, visible=False)
                break

def restoreUIElements():
    global hiddenElements
    for e in hiddenElements:
        #print "restoring... " + e
        cmds.control(e, e=True, visible=True)
    hiddenElements = []

def hideGrid():
    global gridVisible
    gridVisible = cmds.optionVar(q="showGrid")
    cmds.grid(toggle=0)

def restoreGrid():
    global gridVisible
    cmds.grid(toggle=gridVisible)



"""
util
Created by Jonathan deWerd.
"""
import inspect
from traceback import extract_stack
from random import Random
import string
import sys
import os 

import platform
from itertools import dropwhile, starmap
prng = Random()

# qtWrapImport will try using each framework listed
# in the qtFramework list until it finds one that works.
# At that point, qtFramework becomes a string indicating
# the framework that will thereafer be used to load qt classes.
#  Dummy   Uses dummy Qt classes (used by the model when loaded as a module)
#  PyQt    Tries to load Qt classes from PyQt4
#  PySide  Tries to load Qt classes
# The list is defined to consist only of the Dummy qt framework if util is the
# only module that gets loaded. The main.py of applications actually using qt
# need to redefine qtFramework to include PyQt and PySide.

qtFrameworkList = ['PyQt', 'Dummy']
chosenQtFramework = None
# def qtWrapImport(name, globaldict, fromlist):
#     """
#     special function that allows for the import of PySide or PyQt modules
#     as available
# 
#     name is the name of the Qt top level class such as QtCore, or QtGui
# 
#     globaldict is a the module level global namespace dictionary returned from
#     calling the globals() method
# 
#     fromlist is a list of subclasses such as [QFont, QColor], or [QRectF]
#     """
#     global qtWrapFramework  # This method is a stub. It gets swapped out for
#     global qtFrameworkList  # a framework-specific version when called
#     for trialFmwk in qtFrameworkList:
#         if trialFmwk == 'Dummy':
#             qtWrapImport = qtWrapImportFromDummy
#             return qtWrapImport(name, globaldict, fromlist)
#         elif trialFmwk == 'PyQt':
#             try:
#                 import PyQt4
#                 chosenQtFramework = 'PyQt'
#                 qtWrapImport = qtWrapImportFromPyQt
#                 return qtWrapImport(name, globaldict, fromlist)
#             except ImportError:
#                 pass
#         elif trialFmwk == 'PySide':
#             try:
#                 import PySide
#                 chosenQtFramework = 'PySide'
#                 qtWrapImport = qtWrapImportFromPySide
#                 return qtWrapImport(name, globaldict, fromlist)
#             except ImportError:
#                 pass
#         else:
#             raise NameError('Illegal qt framework %s'%trialFmwk)
#     assert(False)  # Have not found a suitable qt framework


def qtWrapImport(name, globaldict, fromlist):
    """
    special function that allows for the import of PySide or PyQt modules
    as available

    name is the name of the Qt top level class such as QtCore, or QtGui

    globaldict is a the module level global namespace dictionary returned from
    calling the globals() method

    fromlist is a list of subclasses such as [QFont, QColor], or [QRectF]
    """
    global qtWrapFramework  # This method is a stub. It gets swapped out for
    global qtFrameworkList  # a framework-specific version when called
    for trialFmwk in qtFrameworkList:
        if trialFmwk == 'PyQt':
            try:
                import PyQt6
                chosenQtFramework = 'PyQt'
                qtWrapImport = qtWrapImportFromPyQt
                return qtWrapImport(name, globaldict, fromlist)
            except ImportError:
                pass
        elif trialFmwk == 'PySide':
            try:
                import PySide
                chosenQtFramework = 'PySide'
                qtWrapImport = qtWrapImportFromPySide
                return qtWrapImport(name, globaldict, fromlist)
            except ImportError:
                pass
        elif trialFmwk == 'Dummy':
            qtWrapImport = qtWrapImportFromDummy
            return qtWrapImport(name, globaldict, fromlist)
        else:
            raise NameError('Illegal qt framework %s'%trialFmwk)
    assert(False)  # Have not found a suitable qt framework


def qtWrapImportFromDummy(name, globaldict, fromlist):
    modName = 'dummyqt.%s'%(name)
    imports = __import__(modName, globaldict, locals(), fromlist, 0)
    canary = object()
    for k in fromlist:
        binding = getattr(imports, k, canary)
        if binding == canary:
            raise KeyError("Couldn't import key '%s' from module '%s'"%(k, modName))
        globaldict[k] = binding


def qtWrapImportFromPyQt(name, globaldict, fromlist):
    modName = 'PyQt6.%s'%(name)
    imports = __import__(modName, globaldict, locals(), fromlist, 0)
    canary = object()
    for key in fromlist:
        binding = getattr(imports, key, canary)
        if binding == canary:
            raise KeyError("Couldn't import key '%s' from module '%s'"%(key, modName))
        globaldict[key] = binding


def qtWrapImportFromPySide(name, globaldict, fromlist):
    modName = 'PySide.%s'%(name)
    imports = __import__(modName, globaldict, locals(), fromlist, -1)
    canary = object()
    for key in fromlist:
        if key in ('pyqtSignal', 'pyqtSlot', 'QString', 'QStringList'):
            if key == 'pyqtSignal':
                globaldict[key] = getattr(imports, 'Signal') 
            elif key == 'pyqtSlot':
                globaldict[key] = getattr(imports, 'Slot')
            elif key == 'QString':
                globaldict[key] = str
            elif key == 'QStringList':
                globaldict[key] = list
        else:
            binding = getattr(_temp, k, canary)
            if binding == canary:
                raise KeyError("Couldn't import key '%s' from module '%s'"%(key, modName))
            globaldict[key] = binding


def clamp(x, minX, maxX):
    if x < minX:
        return minX
    elif x > maxX:
        return maxX
    else:
        return x


def overlap(x,y, a,b):
    """
    finds the overlap of the range x to y in a to b
    assumes an overlap exists, i.e.
        y >= a and b >= x
    """
    c = clamp(x, a, b)
    d = clamp(y, a, b)
    return c, d
# end def


def trace(n):
    """Returns a stack trace n frames deep"""
    s = extract_stack()
    frames = []
    for f in s[-n-1:-1]:
        # f is a stack frame like
        # ('/path/script.py', 42, 'funcname', 'current = line - of / code')
        frames.append( (os.path.basename(f[0])+':%i'%f[1])+'(%s)'%f[2] )
    return " > ".join(frames)


def defineEventForwardingMethodsForClass(classObj, forwardedEventSuffix, eventNames):
    """Automatically defines methods of the form eventName0Event(self, event) on
    classObj that call self.activeTool().eventName0ForwardedEventSuffix(self, event).
    Note that self here is the 2nd argument of eventName0ForwardedEventSuffix which 
    will be defined with 3 arguments, the first of which will implicitly be the
    activeTool(). If self.activeTool() does not implement eventName0ForwardedEventSuffix,
    no error is raised.
    """
    qtWrapImport('QtGui', globals(), [ 'QGraphicsItem', 'QMouseEvent',\
                                       'QGraphicsSceneMouseEvent'])
    for evName in eventNames:
        delegateMethodName = evName + forwardedEventSuffix
        eventMethodName = evName + 'Event'
        defaultMethodName = evName + 'Default'
        forwardDisablingPropertyName = delegateMethodName + 'Unused'

        def makeTemplateMethod(eventMethodName, delegateMethodName):
            def templateMethod(self, event):
                # see if the event also exists
                defaultMethod = getattr(classObj,defaultMethodName,None)
                if defaultMethod != None:
                    defaultMethod(self,event)

                activeTool = self.activeTool()
                if activeTool and not getattr(activeTool, forwardDisablingPropertyName, False):
                    delegateMethod = getattr(activeTool, delegateMethodName, None)
                    if delegateMethod:
                        delegateMethod(self, event)
                    else:
                        superMethod = getattr(QGraphicsItem, eventMethodName)
                        excludeSuperCallBecauseOfTypeIntolerance = \
                                 isinstance(event, QMouseEvent) and\
                                 not isinstance(event, )
                        if not excludeSuperCallBecauseOfTypeIntolerance:
                            superMethod(self, event)
                else:
                    superMethod = getattr(QGraphicsItem, eventMethodName)
                    superMethod(self, event)
            return templateMethod
        eventHandler = makeTemplateMethod(eventMethodName, delegateMethodName)
        setattr(classObj, eventMethodName, eventHandler)


def strToDna(seqStr):
    """Returns str having been reduced to capital ACTG."""
    return "".join([c for c in seqStr if c in 'ACGTacgt']).upper()


complement = str.maketrans('ACGTacgt','TGCATGCA')


def rcomp(seqStr):
    """Returns the reverse complement of the sequence in seqStr."""
    return seqStr.translate(complement)[::-1]


def comp(seqStr):
    """Returns the complement of the sequence in seqStr."""
    return seqStr.translate(complement)


whitetoQ = str.maketrans(' ','?')


def markwhite(seqStr):
    return seqStr.translate(whitetoQ)


def nowhite(seqStr):
    """Gets rid of whitespace in a string."""
    return ''.join([c for c in seqStr if c in string.letters])


nearest=lambda a,l:min(l,key=lambda x:abs(x-a))


def isWindows():
    if platform.system() == 'Windows':
        return True
    else:
        return False


def isMac():
    try:
        return platform.system() == 'Darwin'
    except:
        return os.path.exists('/System/Library/CoreServices/Finder.app')


def isLinux():
    if platform.system() == 'Linux':
        return True
    else:
        return False


def methodName():
    """Returns string containing name of the calling method."""
    return inspect.stack()[1][3]


def starmapExec(f, tupleIter):
    """
    takes a function f and * starmaps the list but drops the results
    """
    list(dropwhile(lambda x: True, starmap(f, tupleIter)))
# end def


def execCommandList(modelObject, commands, desc=None, useUndoStack=True):
    """
    This is a wrapper for performing QUndoCommands, meant to ensure
    uniform handling of the undoStack and macro descriptions.

    When using the undoStack, commands are pushed onto self.undoStack()
    as part of a macro with description desc. Otherwise, command redo
    methods are called directly.
    """
    if useUndoStack:
        undoStackId = str(id(modelObject.undoStack()))[-4:]
        # print "<QUndoStack %s> %s" % (undoStackId, desc)
        modelObject.undoStack().beginMacro(desc)
        for c in commands:
            modelObject.undoStack().push(c)
        modelObject.undoStack().endMacro()
    else:
        # print "<NoUndoStack> %s" % (desc)
        for c in commands:
            c.redo()
# end def


def beginSuperMacro(modelObject, desc=None):
    """
    SuperMacros can be used to nest multiple command lists.

    Normally execCommandList macros all the commands in a list.
    In some cases, multiple command lists need to be executed separately
    because of dependency issues. (e.g. in part.autoStaple, strands
    must be completely 1. created and 2. split before 3. xover installation.)
    """
    modelObject.undoStack().beginMacro(desc)
# end def


def endSuperMacro(modelObject):
    """Ends a SuperMacro. Should be called after beginSuperMacro."""
    modelObject.undoStack().endMacro()
# end def


def findChild(self):
    """
    When called when self isa QGraphicsItem, iterates through self's
    childItems(), placing a red rectangle (a sibling of self) around
    each item in sequence (press return to move between items). Since
    the index of each child item is displayed as it is highlighted,
    one can use findChild() to quickly get a reference to one of self's
    children. At each step, one can type a command letter before
    hitting return. The command will apply to the current child.
    Command Letter:     Action:
    <return>            Advance to next child
    s<return>           Show current child
    S<return>           Show current child, hide siblings
    h<return>           Hide current child
    r<return>           return current child
    """
    qtWrapImport('QtGui', globals(), ['QGraphicsRectItem', 'QPen'])
    qtWrapImport('QtCore', globals(), ['Qt'])
    children = self.childItems()
    parent = self.parentItem()
    childVisibility = [(child, child.isVisible()) for child in children]
    for n in range(len(children)):
        child = children[n]
        print("Highlighting %s.childItems()[%i] = %s"%(self, n, child))
        childBR = child.mapToItem(parent, child.boundingRect())
        childBR = childBR.boundingRect()  # xform gives us a QPolygonF
        debugHighlighter = QGraphicsRectItem(childBR, parent)
        debugHighlighter.setPen(QPen(Qt.GlobalColor.red))
        debugHighlighter.setZValue(9001)
        while True:
            # wait for return to be pressed while spinning the event loop.
            # also process single-character commands.
            command = input()
            if command == 's':    # Show current child
                child.show()
            elif command == 'h':  # Hde current child
                child.hide()
            elif command == 'S':  # Show only current child
                for c in children:
                    c.hide()
                child.show()
            elif command == 'r':  # Return current child
                for child, wasVisible in childVisibility:
                    child.setVisible(wasVisible)
                return child
            else:
                break
        debugHighlighter.scene().removeItem(debugHighlighter)
        for child, wasVisible in childVisibility:
            child.setVisible(wasVisible)



# From https://stackoverflow.com/questions/11130156/
# Define a context manager to suppress stdout and stderr.
class suppress_stdout_stderr(object):
    '''
    A context manager for doing a "deep suppression" of stdout and stderr in 
    Python, i.e. will suppress all print, even if the print originates in a 
    compiled C/Fortran sub-function.
       This will not suppress raised exceptions, since exceptions are printed
    to stderr just before a script exits, and after the context manager has
    exited (at least, I think that is why it lets exceptions through).      

    '''
    def __init__(self):
        # Open a pair of null files
        self.null_fds =  [os.open(os.devnull,os.O_RDWR) for x in range(2)]
        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

    def __enter__(self):
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0],1)
        os.dup2(self.null_fds[1],2)

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0],1)
        os.dup2(self.save_fds[1],2)
        # Close all file descriptors
        for fd in self.null_fds + self.save_fds:
            os.close(fd)



"""
styles.py

Created by Shawn on 2010-06-15.
"""

import cadnano2.util as util
# import Qt stuff into the module namespace with PySide, PyQt4 independence
util.qtWrapImport('QtGui', globals(), [ 'QColor', 'QFont', 'QFontMetricsF'])

# Slice Sizing
SLICE_HELIX_RADIUS = 15
SLICE_HELIX_STROKE_WIDTH = 0.5
SLICE_HELIX_HILIGHT_WIDTH = 2.5
SLICE_HELIX_MOD_HILIGHT_WIDTH = 1
HONEYCOMB_PART_MAXROWS = 30
HONEYCOMB_PART_MAXCOLS = 32
HONEYCOMB_PART_MAXSTEPS = 2
SQUARE_PART_MAXROWS = 50
SQUARE_PART_MAXCOLS = 50
SQUARE_PART_MAXSTEPS = 2

# Slice Colors
bluefill = QColor(153, 204, 255)  # 99ccff
bluestroke = QColor(0, 102, 204)  # 0066cc
bluishstroke = QColor(0, 182, 250)  # 
orangefill = QColor(255, 204, 153)  # ffcc99
orangestroke = QColor(204, 102, 51)  # cc6633
lightorangefill = QColor(255, 234, 183)
lightorangestroke = QColor(234, 132, 81)
grayfill = QColor(238, 238, 238)  # eeeeee (was a1a1a1)
graystroke = QColor(102, 102, 102)  # 666666 (was 424242)

# Path Sizing
VIRTUALHELIXHANDLEITEM_RADIUS = 30
VIRTUALHELIXHANDLEITEM_STROKE_WIDTH = 2
PATH_BASE_WIDTH = 20  # used to size bases (grid squares, handles, etc)
PATH_HELIX_HEIGHT = 2 * PATH_BASE_WIDTH  # staple + scaffold
PATH_HELIX_PADDING = 50 # gap between PathHelix objects in path view
PATH_GRID_STROKE_WIDTH = 0.5
SLICE_HANDLE_STROKE_WIDTH = 1
PATH_STRAND_STROKE_WIDTH = 3
PATH_STRAND_HIGHLIGHT_STROKE_WIDTH = 8
PATH_SELECTBOX_STROKE_WIDTH = 1.5
PCH_BORDER_PADDING = 1
PATH_BASE_HL_STROKE_WIDTH = 2  # PathTool highlight box
MINOR_GRID_STROKE_WIDTH = 0.5
MAJOR_GRID_STROKE_WIDTH = 0.5
oligoLenBelowWhichHighlight = 20
oligoLenAboveWhichHighlight = 49

# Path Drawing
PATH_XOVER_LINE_SCALE_X = 0.035
PATH_XOVER_LINE_SCALE_Y = 0.035

# Path Colors
scaffold_bkg_fill = QColor(230, 230, 230)
activeslicehandlefill = QColor(255, 204, 153, 128)  # ffcc99
activeslicehandlestroke = QColor(204, 102, 51, 128)  # cc6633
minorgridstroke = QColor(204, 204, 204)  # 999999
majorgridstroke = QColor(153, 153, 153)  # 333333
scafstroke = QColor(0, 102, 204)  # 0066cc
handlefill = QColor(0, 102, 204)  # 0066cc
pxi_scaf_stroke = QColor(0, 102, 204, 153)
pxi_stap_stroke = QColor(204, 0, 0, 153)
pxi_disab_stroke = QColor(204, 204, 204, 255)
redstroke = QColor(204, 0, 0)
erasefill = QColor(204, 0, 0, 63)
forcefill = QColor(0, 255, 255, 63)
breakfill = QColor(204, 0, 0, 255)
colorbox_fill = QColor(204, 0, 0)
colorbox_stroke = QColor(102, 102, 102)
stapColors = [QColor(204, 0, 0),
              QColor(247, 67, 8),
              QColor(247, 147, 30),
              QColor(170, 170, 0),
              QColor(87, 187, 0),
              QColor(0, 114, 0),
              QColor(3, 182, 162),
              QColor(23, 0, 222),
              QColor(115, 0, 222),
              QColor(184, 5, 108),
              QColor(51, 51, 51),
              QColor(136, 136, 136)]
scafColors = [QColor(0, 102, 204)]
              # QColor(64, 138, 212),
              # QColor(0, 38, 76),
              # QColor(23, 50, 76),
              # QColor(0, 76, 153)]
DEFAULT_STAP_COLOR = "#888888"
DEFAULT_SCAF_COLOR = "#0066cc"
selected_color = QColor(255, 51, 51)

# brightColors = [QColor() for i in range(10)]
# for i in range(len(brightColors)):
#     brightColors[i].setHsvF(i/12.0, 1.0, 1.0)
# bright_palette = Palette(brightColors)
# cadnn1_palette = Palette(cadnn1Colors)
# default_palette = cadnn1_palette

# Loop/Insertion path details
INSERTWIDTH = 2
SKIPWIDTH = 2

# Add Sequence Tool
INVALID_DNA_COLOR = QColor(255, 224, 224)
UNDERLINE_INVALID_DNA = True

# Additional Prefs
PREF_AUTOSCAF_INDEX = 0
PREF_STARTUP_TOOL_INDEX = 0
PREF_ZOOM_SPEED = 20#50
PREF_ZOOM_AFTER_HELIX_ADD = True


#Z values
#bottom
ZACTIVESLICEHANDLE = 10
ZPATHHELIXGROUP = 20
ZPATHHELIX = 30
ZPATHSELECTION = 40
ZSLICEHELIX = 50
ZDESELECTOR = 60
ZFOCUSRING = 70
ZPREXOVERHANDLE = 80
ZXOVERITEM = 90
ZBREAKPOINTHANDLE = 100
ZSKIPHANDLE = 110
ZBREAKITEM = 120
ZPATHTOOL = 130
ZSTRANDITEM = 140
ZENDPOINTITEM = 150
ZINSERTHANDLE = 160
ZPARTITEM = 200
#top

# sequence stuff
if hasattr(QFont, 'dummy'):
    SEQUENCEFONT = None
    SEQUENCEFONTH = 15
    SEQUENCEFONTCHARWIDTH = 12
    SEQUENCEFONTCHARHEIGHT = 12
    SEQUENCEFONTEXTRAWIDTH = 3
    SEQUENCETEXTXCENTERINGOFFSET = 0
else:
    SEQUENCEFONT = QFont("Monaco")
    if hasattr(QFont, 'Monospace'):
        SEQUENCEFONT.setStyleHint(QFont.Monospace)
    SEQUENCEFONT.setFixedPitch(True)
    SEQUENCEFONTH = PATH_BASE_WIDTH // 3
    SEQUENCEFONT.setPixelSize(SEQUENCEFONTH)
    SEQUENCEFONTMETRICS = QFontMetricsF(SEQUENCEFONT)
    SEQUENCEFONTCHARWIDTH = SEQUENCEFONTMETRICS.horizontalAdvance('A')
    SEQUENCEFONTCHARHEIGHT = SEQUENCEFONTMETRICS.height()
    SEQUENCEFONTEXTRAWIDTH = PATH_BASE_WIDTH - SEQUENCEFONTCHARWIDTH
    SEQUENCEFONT.setLetterSpacing(QFont.SpacingType.AbsoluteSpacing,
                                 SEQUENCEFONTEXTRAWIDTH)
    SEQUENCETEXTXCENTERINGOFFSET = SEQUENCEFONTEXTRAWIDTH / 4.
    SEQUENCETEXTYCENTERINGOFFSET = PATH_BASE_WIDTH * 0.6


if util.isMac():
    thefont = "Times"
    thefont = "Arial"
    thefontsize = 10
    XOVER_LABEL_FONT = QFont(thefont, thefontsize, QFont.Weight.Bold)
elif util.isWindows():
    thefont = "Segoe UI"
    thefont = "Calibri"
    thefont = "Arial"
    thefontsize = 9
    XOVER_LABEL_FONT = QFont(thefont, thefontsize, QFont.Weight.Bold)
else: # linux
    thefont = "DejaVu Sans"
    thefontsize = 9
    XOVER_LABEL_FONT = QFont(thefont, thefontsize, QFont.Weight.Bold)
 
SLICE_NUM_FONT = QFont(thefont, 10, QFont.Weight.Bold)
VIRTUALHELIXHANDLEITEM_FONT = QFont(thefont, 3*thefontsize, QFont.Weight.Bold)
XOVER_LABEL_COLOR = QColor(0,0,0) 

# Overwrite for Maya
# majorgridstroke = QColor(255, 255, 255)  # ffffff for maya

